//-------------------------------------------------------------------------------------------------
// <copyright file="DriveCheck.cpp" company="Microsoft">
//    Copyright (c) Microsoft Corporation.  All rights reserved.
//    
//    The use and distribution terms for this software are covered by the
//    Common Public License 1.0 (http://opensource.org/licenses/cpl.php)
//    which can be found in the file CPL.TXT at the root of this distribution.
//    By using this software in any fashion, you are agreeing to be bound by
//    the terms of this license.
//    
//    You must not remove this notice, or any other, from this software.
// </copyright>
// 
// <summary>
//    Validate install drive
// </summary>
//-------------------------------------------------------------------------------------------------

#include "precomp.h"

static HRESULT PathIsRemote(__in LPCWSTR pTargetFolder, __inout BOOL* fPathRemote);
static HRESULT PathIsRemovable(__in LPCWSTR pTargetFolder, __inout BOOL* fPathRemovable);

/********************************************************************
 ValidatePath - Custom Action entry point

********************************************************************/
UINT __stdcall ValidatePath(MSIHANDLE hInstall)
{
    HRESULT hr = S_OK;
    
    LPWSTR pwszWixUIDir = NULL;
    LPWSTR pwszInstallPath = NULL;
    BOOL fInstallPathIsRemote = TRUE;
    BOOL fInstallPathIsRemoveable = TRUE;
    
    hr = WcaInitialize(hInstall, "ValidatePath");
    ExitOnFailure(hr, "failed to initialize");

    hr = WcaGetProperty(L"WIXUI_INSTALLDIR", &pwszWixUIDir);
    ExitOnFailure(hr, "failed to get WixUI Installation Directory");
    
    hr = WcaGetProperty(pwszWixUIDir, &pwszInstallPath);
    ExitOnFailure(hr, "failed to get Installation Directory");
    
    hr = PathIsRemote(pwszInstallPath, &fInstallPathIsRemote);
    if(FAILED(hr))
    {
        TraceError(hr, "Unable to determine if path is remote");
        //reset HR, as we need to continue and find out if is a UNC path
        hr = S_OK;
    }
    
    hr = PathIsRemovable(pwszInstallPath, &fInstallPathIsRemoveable);
    if(FAILED(hr))
    {
        TraceError(hr, "Unable to determine if path is remote");
        //reset HR, as we need to continue and find out if is a UNC path
        hr = S_OK;
    }

    // If the path does not point to a network drive, mapped drive, relative path, or removable drive,
    // then set WIXUI_INSTALLDIR_VALID to "1" otherwise set it to 0
    if (!PathIsUNCW(pwszInstallPath) && !fInstallPathIsRemote && !fInstallPathIsRemoveable)
    {
        // path is valid
        hr = WcaSetProperty(L"WIXUI_INSTALLDIR_VALID",  L"1");
        ExitOnFailure(hr, "failed to set WIXUI_INSTALLDIR_VALID");
    }
    else
    {
        // path is invalid
        hr = WcaSetProperty(L"WIXUI_INSTALLDIR_VALID",  L"0");
        ExitOnFailure(hr, "failed to set WIXUI_INSTALLDIR_VALID");
    }
    
LExit:
    ReleaseStr(pwszInstallPath);
    ReleaseStr(pwszWixUIDir);
    
    if(SUCCEEDED(hr))
    {
        return WcaFinalize(ERROR_SUCCESS);
    }
    else
    {
        return WcaFinalize(ERROR_BAD_PATHNAME);
    }
}

/********************************************************************
 PathIsRemote - helper function for ValidatePath

********************************************************************/
static HRESULT PathIsRemote(__in LPCWSTR pTargetFolder, __inout BOOL* fPathRemote)
{
    HRESULT hr = S_OK;
    LPWSTR pStrippedTargetFolder = NULL;

    hr = StrAllocString(&pStrippedTargetFolder, pTargetFolder, 0);
    
    // Terminate the path at the root
    if(!::PathStripToRootW(pStrippedTargetFolder))
    {
        hr = HRESULT_FROM_WIN32(ERROR_INVALID_DRIVE);
        ExitOnFailure(hr, "failed to parse target folder");    
    }
    
    UINT uResult = GetDriveTypeW(pStrippedTargetFolder);
    
    *fPathRemote = (DRIVE_REMOTE == uResult) ;

LExit:
    ReleaseStr(pStrippedTargetFolder);

    return hr;        
}

/********************************************************************
 PathIsRemovable - helper function for ValidatePath

********************************************************************/
static HRESULT PathIsRemovable(__in LPCWSTR pTargetFolder, __inout BOOL* fPathRemovable)
{
    HRESULT hr = S_OK;
    LPWSTR pStrippedTargetFolder = NULL;

    hr = StrAllocString(&pStrippedTargetFolder, pTargetFolder, 0);
    
    // Terminate the path at the root
    if(!::PathStripToRootW(pStrippedTargetFolder))
    {
        hr = HRESULT_FROM_WIN32(ERROR_INVALID_DRIVE);
        ExitOnFailure(hr, "failed to parse target folder");    
    }
    
    UINT uResult = GetDriveTypeW(pStrippedTargetFolder);
    
    *fPathRemovable = ((DRIVE_CDROM == uResult) || (DRIVE_REMOVABLE == uResult) || (DRIVE_RAMDISK == uResult) || (DRIVE_UNKNOWN == uResult));

LExit:
    ReleaseStr(pStrippedTargetFolder);

    return hr;
}
