//-------------------------------------------------------------------------------------------------
// <copyright file="UtilHeatExtension.cs" company="Microsoft">
//    Copyright (c) Microsoft Corporation.  All rights reserved.
//    
//    The use and distribution terms for this software are covered by the
//    Common Public License 1.0 (http://opensource.org/licenses/cpl.php)
//    which can be found in the file CPL.TXT at the root of this distribution.
//    By using this software in any fashion, you are agreeing to be bound by
//    the terms of this license.
//    
//    You must not remove this notice, or any other, from this software.
// </copyright>
// 
// <summary>
// A utility heat extension for the Windows Installer XML Toolset Harvester application.
// </summary>
//-------------------------------------------------------------------------------------------------

namespace Microsoft.Tools.WindowsInstallerXml.Extensions
{
    using System;
    using System.Collections.Generic;
    using System.Collections.Specialized;
    using System.IO;
    using Microsoft.Tools.WindowsInstallerXml.Tools;

    using Wix = Microsoft.Tools.WindowsInstallerXml.Serialize;

    /// <summary>
    /// A utility heat extension for the Windows Installer XML Toolset Harvester application.
    /// </summary>
    public sealed class UtilHeatExtension : HeatExtension
    {
        /// <summary>
        /// Gets the supported command line types for this extension.
        /// </summary>
        /// <value>The supported command line types for this extension.</value>
        public override HeatCommandLineOption[] CommandLineTypes
        {
            get
            {
                return new HeatCommandLineOption[]
                {
                    new HeatCommandLineOption("dir", "harvest a directory"),
                    new HeatCommandLineOption("file", "harvest a file"),
                    new HeatCommandLineOption("-ag", "autogenerate component guids at compile time"),
                    new HeatCommandLineOption("-gg", "generate guids now"),
                    new HeatCommandLineOption("-g1", "generated guids are not in brackets"),
                    new HeatCommandLineOption("-sfrag", "suppress fragments"),
                    new HeatCommandLineOption("-ke", "keep empty directories"),
                    new HeatCommandLineOption("-scom", "suppress COM elements"),
                    new HeatCommandLineOption("-svb6", "suppress VB6 COM elements"),
                    new HeatCommandLineOption("-sreg", "suppress registry harvesting"),
                    new HeatCommandLineOption("-suid", "suppress unique identifiers for files, components, & directories"),
                    new HeatCommandLineOption("-t:<xsl>", "transform harvested output with XSL file"),
                    new HeatCommandLineOption("-template:", "use template, one of: fragment,module,product"),
                };
            }
        }

        /// <summary>
        /// Parse the command line options for this extension.
        /// </summary>
        /// <param name="type">The active harvester type.</param>
        /// <param name="args">The option arguments.</param>
        public override void ParseOptions(string type, string[] args)
        {
            bool active = false;
            HarvesterExtension harvesterExtension = null;
            bool suppressHarvestingRegistryValues = false;
            UtilFinalizeHarvesterMutator utilFinalizeHarvesterMutator = new UtilFinalizeHarvesterMutator();
            UtilMutator utilMutator = new UtilMutator();
            List<UtilTransformMutator> transformMutators = new List<UtilTransformMutator>();

            // select the harvester
            switch (type)
            {
                case "dir":
                    harvesterExtension = new DirectoryHarvester();
                    active = true;
                    break;
                case "file":
                    harvesterExtension = new FileHarvester();
                    active = true;
                    break;
            }

            // set default settings
            utilMutator.CreateFragments = true;
            utilMutator.SetUniqueIdentifiers = true;

            // parse the options
            foreach (string arg in args)
            {
                if (null == arg || 0 == arg.Length) // skip blank arguments
                {
                    continue;
                }

                if ('-' == arg[0] || '/' == arg[0])
                {
                    string parameter = arg.Substring(1);

                    if ("ag" == parameter)
                    {
                        utilMutator.AutogenerateGuids = true;
                    }
                    else if ("gg" == parameter)
                    {
                        utilMutator.GenerateGuids = true;
                    }
                    else if ("g1" == parameter)
                    {
                        utilMutator.GuidFormat = "D";
                    }
                    else if ("ke" == parameter)
                    {
                        if (harvesterExtension is DirectoryHarvester)
                        {
                            ((DirectoryHarvester)harvesterExtension).KeepEmptyDirectories = true;
                        }
                        else if (active)
                        {
                            // TODO: error message - not applicable to file harvester
                        }
                    }
                    else if ("scom" == parameter)
                    {
                        if (active)
                        {
                            utilFinalizeHarvesterMutator.SuppressCOMElements = true;
                        }
                        else
                        {
                            // TODO: error message - not applicable
                        }
                    }
                    else if ("svb6" == parameter)
                    {
                        if (active)
                        {
                            utilFinalizeHarvesterMutator.SuppressVB6COMElements = true;
                        }
                        else
                        {
                            // TODO: error message - not applicable
                        }
                    }
                    else if ("sfrag" == parameter)
                    {
                        utilMutator.CreateFragments = false;
                    }
                    else if ("sreg" == parameter)
                    {
                        suppressHarvestingRegistryValues = true;
                    }
                    else if ("suid" == parameter)
                    {
                        utilMutator.SetUniqueIdentifiers = false;
                    }
                    else if (parameter.StartsWith("t:", StringComparison.Ordinal))
                    {
                        string xslFile = parameter.Substring(2);

                        if (0 <= xslFile.IndexOf('\"'))
                        {
                            this.Core.OnMessage(WixErrors.PathCannotContainQuote(xslFile));
                            return;
                        }

                        try
                        {
                            xslFile = Path.GetFullPath(xslFile);
                        }
                        catch (Exception e)
                        {
                            this.Core.OnMessage(WixErrors.InvalidCommandLineFileName(xslFile, e.Message));
                            return;
                        }

                        transformMutators.Add(new UtilTransformMutator(xslFile, transformMutators.Count));
                    }
                    else if (parameter.StartsWith("template:", StringComparison.Ordinal))
                    {
                        switch (parameter.Substring(9))
                        {
                            case "fragment":
                                utilMutator.TemplateType = TemplateType.Fragment;
                                break;
                            case "module":
                                utilMutator.TemplateType = TemplateType.Module;
                                break;
                            case "product":
                                utilMutator.TemplateType = TemplateType.Product;
                                break;
                            default:
                                // TODO: error
                                break;
                        }
                    }
                }
            }

            // set the appropriate harvester extension
            if (active)
            {
                this.Core.Harvester.Extension = harvesterExtension;

                if (!suppressHarvestingRegistryValues)
                {
                    this.Core.Mutator.AddExtension(new UtilHarvesterMutator());
                }

                this.Core.Mutator.AddExtension(utilFinalizeHarvesterMutator);
            }

            // set the mutator
            this.Core.Mutator.AddExtension(utilMutator);

            // add the transforms
            foreach (UtilTransformMutator transformMutator in transformMutators)
            {
                this.Core.Mutator.AddExtension(transformMutator);
            }
        }
    }
}
